#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

# Directory name which scenario file is placed
SCENARIO_DIR_NAME = "scenario"

# scenario file name excluding extension
SCENARIO_FILE_NAME = "scenario"

# cliboa project directory path
BASE_DIR = os.getcwd()

# Project directory path. Customization is available
PROJECT_DIR = os.path.join(BASE_DIR, "project")

# Common scenario directory path. Customization is available
COMMON_DIR = os.path.join(BASE_DIR, "common")

# Common scenario directory path. Customization is available
COMMON_SCENARIO_DIR = os.path.join(COMMON_DIR, "scenario")

# the blow paths are appended to sys.path of python
SYSTEM_APPEND_PATHS = [COMMON_SCENARIO_DIR]

# common custom classes to make available
COMMON_CUSTOM_CLASSES = ["sample_step.SampleStep"]

# project congenital classes to make available
PROJECT_CUSTOM_CLASSES = []



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import json
import os
import random
import re
import string
from datetime import datetime

import pandas
from google.cloud import bigquery

from cliboa.scenario.gcp import BaseBigQuery, BaseFirestore, BaseGcs
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.cache import ObjectStore
from cliboa.util.exception import InvalidParameter
from cliboa.util.gcp import BigQuery, Firestore, Gcs, ServiceAccount
from cliboa.util.string import StringUtil


class BigQueryRead(BaseBigQuery):
    """
    Read data from BigQuery and put them into on-memory or export to a file via GCS.
    """

    _RANDOM_STR_LENGTH = 8

    def __init__(self):
        super().__init__()
        self._key = None
        self._bucket = None
        self._dest_dir = None
        self._filename = None
        self._query = None

    def key(self, key):
        self._key = key

    def query(self, query):
        self._query = query

    def bucket(self, bucket):
        self._bucket = bucket

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def filename(self, filename):
        self._filename = filename

    def execute(self, *args):
        super().execute()
        if not self._key and not self._bucket:
            raise InvalidParameter("Specifying either 'key' or 'bucket' is essential.")
        if self._key and self._bucket:
            raise InvalidParameter("Cannot specify both 'key' and 'bucket'.")

        # fetch records and save to on-memory
        if self._key:
            valid = EssentialParameters(self.__class__.__name__, [self._tblname])
            valid()
            self._save_to_cache()
        elif self._bucket:
            valid = EssentialParameters(self.__class__.__name__, [self._dest_dir])
            valid()
            self._save_as_file_via_gcs()

    def _save_to_cache(self):
        self._logger.info("Save data to on memory")
        df = pandas.read_gbq(
            query="SELECT * FROM %s.%s" % (self._dataset, self._tblname)
            if self._query is None
            else self._query,
            dialect="standard",
            location=self._location,
            project_id=self._project_id,
            credentials=ServiceAccount.auth(self._credentials),
        )
        ObjectStore.put(self._key, df)

    def _save_as_file_via_gcs(self):
        self._logger.info("Save data as a file via GCS")
        os.makedirs(self._dest_dir, exist_ok=True)

        ymd_hms = datetime.now().strftime("%Y%m%d%H%M%S%f")
        path = "%s-%s" % (StringUtil().random_str(self._RANDOM_STR_LENGTH), ymd_hms,)
        prefix = "%s/%s/%s" % (self._dataset, self._tblname, path)

        gbq_client = BigQuery.get_bigquery_client(self._credentials)
        if self._dataset and self._tblname:
            table_ref = gbq_client.dataset(self._dataset).table(self._tblname)
        elif self._dataset and not self._tblname:
            tmp_tbl = (
                "tmp_"
                + StringUtil().random_str(self._RANDOM_STR_LENGTH)
                + "_"
                + ymd_hms
            )
            table_ref = gbq_client.dataset(self._dataset).table(tmp_tbl)

        gcs_client = Gcs.get_gcs_client(self._credentials)
        gcs_bucket = gcs_client.bucket(self._bucket)

        # extract job config settings
        ext_job_config = BigQuery.get_extract_job_config()
        ext_job_config.compression = BigQuery.get_compression_type()
        ext = ".csv"
        if self._filename:
            _, ext = os.path.splitext(self._filename)
            support_ext = [".csv", ".json"]
            if ext not in support_ext:
                raise InvalidParameter("%s is not supported as filename." % ext)
        ext_job_config.destination_format = BigQuery.get_destination_format(ext)

        comp_format_and_ext = {"GZIP": ".gz"}
        comp_ext = comp_format_and_ext.get(str(BigQuery.get_compression_type()))
        if self._filename:
            dest_gcs = "gs://%s/%s/%s%s" % (
                self._bucket,
                prefix,
                self._filename,
                comp_ext,
            )
        else:
            dest_gcs = "gs://%s/%s/*%s%s" % (self._bucket, prefix, ext, comp_ext)

        # Execute query.
        if self._query:
            query_job_config = BigQuery.get_query_job_config()
            query_job_config.destination = table_ref
            query_job_config.write_disposition = BigQuery.get_write_disposition()
            query_job = gbq_client.query(
                self._query, location=self._location, job_config=query_job_config
            )
            query_job.result()

        # Extract to GCS
        extract_job = gbq_client.extract_table(
            table_ref, dest_gcs, job_config=ext_job_config, location=self._location
        )
        extract_job.result()

        # Download from gcs
        for blob in gcs_bucket.list_blobs(prefix=prefix):
            dest = os.path.join(self._dest_dir, os.path.basename(blob.name))
            blob.download_to_filename(dest)

        # Cleanup temporary table
        if self._query:
            gbq_client.delete_table(table_ref)

        # Cleanup temporary files
        for blob in gcs_bucket.list_blobs(prefix=prefix):
            blob.delete()


class BigQueryReadCache(BaseBigQuery):
    """
    @deprecated
    Get data from BigQuery and cache them as pandas.dataframe format.

    Use {BigQueryFileDownload} if the result query is estimated to be large.
    """

    def __init__(self):
        super().__init__()
        self._key = None
        self._query = None

    def key(self, key):
        self._key = key

    def query(self, query):
        self._query = query

    def _get_query(self):
        if self._query is None:
            return "SELECT * FROM %s.%s" % (self._dataset, self._tblname)
        else:
            return self._query

    def execute(self, *args):
        super().execute()
        valid = EssentialParameters(self.__class__.__name__, [self._key])
        valid()

        df = pandas.read_gbq(
            query=self._get_query(),
            dialect="standard",
            location=self._location,
            project_id=self._project_id,
            credentials=ServiceAccount.auth(self._credentials),
        )
        ObjectStore.put(self._key, df)


class BigQueryFileDownload(BaseBigQuery):
    """
    @deprecated
    Download query result as a csv file.

    This class saves BigQuery result as a temporary file in GCS, and then download.
    Download file could be a multiple if file size exceed 1GB.
    """

    def __init__(self):
        super().__init__()
        self._bucket = None
        self._dest_dir = None
        self._filename = None

    def bucket(self, bucket):
        self._bucket = bucket

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def filename(self, filename):
        self._filename = filename

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(
            self.__class__.__name__, [self._tblname, self._bucket, self._dest_dir]
        )
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        gbq_client = BigQuery.get_bigquery_client(self._credentials)
        gbq_ref = gbq_client.dataset(self._dataset).table(self._tblname)

        gcs_client = Gcs.get_gcs_client(self._credentials)
        gcs_bucket = gcs_client.bucket(self._bucket)

        ymd_hms = datetime.now().strftime("%Y%m%d%H%M%S%f")
        path = "%s-%s" % ("".join(random.choices(string.ascii_letters, k=8)), ymd_hms)
        prefix = "%s/%s/%s" % (self._dataset, self._tblname, path)

        """
        gsc dir -> gs://{bucket_name}
                       /{dataset_name}/{table_name}
                       /{XXXXXXXX}-{yyyyMMddHHmmssSSS}/*.csv.gz
        """
        if self._filename:
            dest_gcs = "gs://%s/%s/%s*.csv.gz" % (self._bucket, prefix, self._filename)
        else:
            dest_gcs = "gs://%s/%s/*.csv.gz" % (self._bucket, prefix)

        # job config settings
        job_config = bigquery.ExtractJobConfig()
        job_config.compression = bigquery.Compression.GZIP
        job_config.desctination_format = bigquery.DestinationFormat.CSV

        # Execute query.
        job = gbq_client.extract_table(
            gbq_ref, dest_gcs, job_config=job_config, location=self._location
        )
        job.result()

        # Download from gcs
        for blob in gcs_client.list_blobs(gcs_bucket, prefix=prefix):
            dest = os.path.join(self._dest_dir, os.path.basename(blob.name))
            blob.download_to_filename(dest)

        # Cleanup temporary files
        for blob in gcs_client.list_blobs(gcs_bucket, prefix=prefix):
            blob.delete()


class GcsDownload(BaseGcs):
    """
    Download from GCS
    """

    def __init__(self):
        super().__init__()
        self._prefix = None
        self._delimiter = None
        self._src_pattern = None
        self._dest_dir = "."

    def prefix(self, prefix):
        self._prefix = prefix

    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        client = Gcs.get_gcs_client(self._credentials)
        bucket = client.bucket(self._bucket)
        dl_files = []
        for blob in client.list_blobs(
            bucket, prefix=self._prefix, delimiter=self._delimiter
        ):
            r = re.compile(self._src_pattern)
            if not r.fullmatch(blob.name):
                continue
            dl_files.append(blob.name)
            blob.download_to_filename(
                os.path.join(self._dest_dir, os.path.basename(blob.name))
            )

        ObjectStore.put(self._step, dl_files)


class GcsDownloadFileDelete(BaseGcs):
    """
    Delete files downloaded from GCS by using cache
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        dl_files = ObjectStore.get(self._symbol)

        if len(dl_files) > 0:
            self._logger.info("Delete files %s" % dl_files)
            client = self._gcs_client()
            bucket = client.bucket(super().get_step_argument("bucket"))
            for blob in client.list_blobs(
                bucket,
                prefix=super().get_step_argument("prefix"),
                delimiter=super().get_step_argument("delimiter"),
            ):
                for dl_f in dl_files:
                    if dl_f == blob.name:
                        blob.delete()
                        break
        else:
            self._logger.info("No files to delete.")


class FirestoreDownloadDocument(BaseFirestore):
    """
    Download a document from Firestore
    """

    def __init__(self):
        super().__init__()
        self._dest_dir = None

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(
            self.__class__.__name__, [self._collection, self._document, self._dest_dir]
        )
        valid()

        firestore_client = Firestore.get_firestore_client(self._credentials)
        ref = firestore_client.document(self._collection, self._document)
        doc = ref.get()

        with open(os.path.join(self._dest_dir, doc.id), mode="wt") as f:
            f.write(json.dumps(doc.to_dict()))



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import ast
import csv

from cliboa.scenario.base import BaseStep
from cliboa.scenario.validator import IOOutput


class FileWrite(BaseStep):
    """
    Load file to server
    """

    def __init__(self):
        super().__init__()
        self._src_dir = None
        self._src_pattern = None
        self._dest_path = None
        self._encoding = "utf-8"
        self._mode = "a"

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_path(self, dest_path):
        self._dest_path = dest_path

    def encoding(self, encoding):
        self._encoding = encoding

    def mode(self, mode):
        self._mode = mode


class CsvWrite(FileWrite):
    """
    Write data fetched by io: inptu to csv file
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        super().execute()
        input_valid = IOOutput(self._io)
        input_valid()

        with open(self._s.cache_file, "r", encoding=self._encoding) as i, open(
            self._dest_path, self._mode, encoding=self._encoding
        ) as o:
            writer = csv.writer(o, quoting=csv.QUOTE_ALL)

            # write csv header
            head_dict = ast.literal_eval(i.readline())
            header = []
            for k in head_dict.keys():
                header.append(k)
            writer.writerow(header)

            # write as csv per one line
            i.seek(0)
            for l_str in i:
                l_dict = ast.literal_eval(l_str)
                contents = []
                for k, v in l_dict.items():
                    contents.append(v)
                writer.writerow(contents)
        self._s.remove()



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
from abc import abstractmethod

import yaml

from cliboa.core.validator import EssentialKeys, ScenarioYamlKey, ScenarioYamlType  # noqa
from cliboa.util.lisboa_log import LisboaLog


class ScenarioParser(object):
    """
    Base class of scnario file parser
    """

    def __init__(self, pj_scenario_file, cmn_scenario_file):
        self._logger = LisboaLog.get_logger(__name__)
        self._pj_scenario_file = pj_scenario_file
        self._cmn_scenario_file = cmn_scenario_file

    @abstractmethod
    def parse(self):
        """
        Parse scenario file
        """


class YamlScenarioParser(ScenarioParser):
    """
    scenario.yml parser
    """

    def parse(self):
        self._logger.info("Start to parse scenario.yml")
        pj_f = open(self._pj_scenario_file, "r")
        pj_yaml_dict = yaml.safe_load(pj_f)

        self.__valid_scenario_yaml(pj_yaml_dict)
        exists_cmn_scenario_file = os.path.isfile(self._cmn_scenario_file)
        if exists_cmn_scenario_file:
            cmn_f = open(self._cmn_scenario_file, "r")
            cmn_yaml_dict = yaml.safe_load(cmn_f)
            self.__valid_scenario_yaml(cmn_yaml_dict)
            self.__exists_ess_keys(pj_yaml_dict["scenario"])
            self.__exists_ess_keys(cmn_yaml_dict["scenario"])
            yaml_list = self.__merge_scenario_yaml(
                pj_yaml_dict["scenario"], cmn_yaml_dict["scenario"]
            )
            cmn_f.close()
        else:
            yaml_list = pj_yaml_dict.get("scenario")

        self._logger.info("Finish to parse scenario.yml")
        pj_f.close()
        return yaml_list

    def __valid_scenario_yaml(self, yaml_dict):
        """
        validate instance type and essential key in scenario.yml
        """
        for scenario_yaml_valid in ["ScenarioYamlType", "ScenarioYamlKey"]:
            cls_name = globals()[scenario_yaml_valid](yaml_dict)
            cls_name()

    def __merge_scenario_yaml(self, pj_yaml_list, cmn_yaml_list):
        """
        Merge project scenrio.yml and common scenairo.yml.
        If the same class specification exists,
        scenario.yml of projet is taken priority.
        """
        for pj_yaml_dict in pj_yaml_list:
            # If same class exists, merge arguments
            if pj_yaml_dict.get("parallel"):
                for row in pj_yaml_dict.get("parallel"):
                    self._merge(row, cmn_yaml_list)
            else:
                self._merge(pj_yaml_dict, cmn_yaml_list)

        return pj_yaml_list

    def _merge(self, pj_yaml_dict, cmn_yaml_list):
        cmn_yaml_dict_in_list = [
            d for d in cmn_yaml_list if d.get("class") == pj_yaml_dict.get("class")
        ]
        if not cmn_yaml_dict_in_list:
            return

        pj_cls_attrs = pj_yaml_dict.get("arguments", "")
        cmn_cls_attrs = cmn_yaml_dict_in_list[0].get("arguments")

        # Merge arguments
        if pj_cls_attrs and cmn_cls_attrs:
            pj_cls_attrs = dict(cmn_cls_attrs, **pj_cls_attrs)
        elif not pj_cls_attrs and cmn_cls_attrs:
            pj_cls_attrs = cmn_cls_attrs
        pj_yaml_dict["arguments"] = pj_cls_attrs

    def __exists_ess_keys(self, scenario_yaml_list):
        """
        Check if the essential keys exist in scenario.yml
        """
        valid = EssentialKeys(scenario_yaml_list)
        valid()


class JsonScenarioParser:
    """
    TODO: implement in the future
    """



class Helper(object):
    @staticmethod
    def set_property(cls, method, value):
        """
        This method allows you to set a value
        to the class with either method directly or via property setter.
        Either way, the method must be implemented
        to set the value for the class parameter like below.

        -- eg1 --
        class Foo():
            def __init__(self):
                self._bar = None

            def bar(bar):
                self._bar = bar

        -- eg2 --
        class Foo2():
            def __init__(self):
                self._bar = None

            @property
            def bar(self):
                return self._bar

            @bar.setter
            def bar(self, bar):
                self._bar = bar

        Arguments
            cls (Class): Class instance
            method (str): method name
            value (any): value to set
        """
        attr = getattr(type(cls), method, None)
        if isinstance(attr, property):
            # Set property via property.setter
            setattr(cls, method, value)
        else:
            # Set property directly by method
            call = getattr(cls, method, None)
            if call is not None:
                call(value)



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from abc import ABC, abstractmethod
from time import sleep

import requests
from requests.exceptions import HTTPError

from cliboa.scenario.validator import EssentialParameters
from cliboa.util.lisboa_log import LisboaLog


class FormAuth(object):
    """
    form authenticadtion
    """

    def __init__(self):
        self.__form_url = None
        self.__form_id = None
        self.__form_password = None

    @property
    def form_url(self):
        return self.__form_url

    @form_url.setter
    def form_url(self, form_url):
        self.__form_url = form_url

    @property
    def form_id(self):
        return self.__form_id

    @form_id.setter
    def form_id(self, form_id):
        self.__form_id = form_id

    @property
    def form_password(self):
        return self.__form_password

    @form_password.setter
    def form_password(self, form_password):
        self.__form_password = form_password

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self.__form_url, self.__form_id, self.__form_password],
        )
        valid()


class BasicAuth(object):
    """
    TBD implement in the future
    """

    def __init__(self):
        super().__init__()
        self.__basic_auth = False

    @property
    def basic_auth(self):
        return self.__basic_auth

    @basic_auth.setter
    def basic_auth(self, basic_auth):
        self.__basic_auth = basic_auth


class Http(ABC):
    """
    Http client abstract class
    """

    def __init__(self, url, dest_path, timeout, retry_cnt, retry_intvl_sec):
        self._logger = LisboaLog.get_logger(__name__)
        self._url = url
        self._dest_path = dest_path
        self._timeout = timeout
        self._retry_cnt = retry_cnt
        self._retry_intvl_sec = retry_intvl_sec

    @abstractmethod
    def execute(self):
        pass


class Download(Http):
    """
    Download the specific file via simple Http GET
    """

    VALID_HTTP_STATUS = 200

    def __init__(self, url, dest_path, timeout, retry_cnt, retry_intvl_sec=10):
        super().__init__(url, dest_path, timeout, retry_cnt, retry_intvl_sec)

    def execute(self):
        self._logger.info("Http GET url: %s" % self._url)
        self._logger.info("local path: %s" % self._dest_path)
        res = None
        for _ in range(self._retry_cnt):
            try:
                res = requests.get(self._url, timeout=self._timeout)
                res.raise_for_status()
                with open(self._dest_path, "wb") as f:
                    f.write(res.content)
                return

            except Exception as e:
                self._logger.warning(e)

            self._logger.warning(
                "Unexpected error occurred during http get. Retry will start in %s"
                % self._retry_intvl_sec
            )
            sleep(self._retry_intvl_sec)

        if res is not None:
            raise HTTPError(
                "Http request failed. HTTP Status code: %s" % res.status_code
            )


class DownloadViaBasicAuth(Http):
    """
    TBD implement in the future
    """

    def __init__(self, id, password):
        self.__id = id
        self.__pass = password

    def execute(self):
        pass



#
# Copyright 2020 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from google.cloud import bigquery, firestore, storage
from google.oauth2 import service_account

from cliboa.scenario.base import BaseStep
from cliboa.scenario.validator import EssentialParameters


class BaseGcp(BaseStep):
    """
    Base class of Gcp usage.
    """

    def __init__(self):
        super().__init__()
        self._project_id = None
        self._credentials = None

    def project_id(self, project_id):
        self._project_id = project_id

    def credentials(self, credentials):
        self._credentials = credentials

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._project_id])
        valid()

    def _auth(self):
        """
        @deprecated
        use ServiceAccount.auth() in util.gcp
        """
        if self._credentials:
            return service_account.Credentials.from_service_account_file(
                self._credentials
            )

    def _bigquery_client(self):
        """
        @deprecated
        use BigQuery.get_bigquery_client() in util.gcp
        """
        if self._credentials:
            return bigquery.Client.from_service_account_json(self._credentials)
        else:
            return bigquery.Client()

    def _gcs_client(self):
        """
        @deprecated
        use Gcs.get_gcs_client() in util.gcp
        """
        if self._credentials:
            return storage.Client.from_service_account_json(self._credentials)
        else:
            return storage.Client()

    def _firestore_client(self):
        """
        @deprecated
        use Firestore.get_firestore_client() in util.gcp
        """
        if self._credentials:
            return firestore.Client.from_service_account_json(self._credentials)
        else:
            return firestore.Client()


class BaseBigQuery(BaseGcp):
    """
    Base class of BigQuery use.

    """

    def __init__(self):
        super().__init__()

        self._dataset = None
        self._tblname = None
        self._location = None

    def dataset(self, dataset):
        self._dataset = dataset

    def tblname(self, tblname):
        self._tblname = tblname

    def location(self, location):
        self._location = location

    def execute(self, *args):
        super().execute()
        valid = EssentialParameters(
            self.__class__.__name__, [self._location, self._dataset]
        )
        valid()


class BaseGcs(BaseGcp):
    """
    Base class of Gcs use.

    """

    def __init__(self):
        super().__init__()
        self._bucket = None

    def bucket(self, bucket):
        self._bucket = bucket

    def execute(self, *args):
        super().execute()
        valid = EssentialParameters(self.__class__.__name__, [self._bucket])
        valid()


class BaseFirestore(BaseGcp):
    """
    Base class of Firebase use.

    """

    def __init__(self):
        super().__init__()
        self._collection = None
        self._document = None

    def collection(self, collection):
        self._collection = collection

    def document(self, document):
        self._document = document
